#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Verification tests for solubility product
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    units as pyunits,
    value,
)

# Import IDAES cores
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
    ConcentrationForm,
)
from idaes.models.unit_models import EquilibriumReactor
from idaes.core import (
    Component,
    EnergyBalanceType,
    FlowsheetBlock,
    LiquidPhase,
    SolidPhase,
)
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

from idaes.models.properties.modular_properties.state_definitions import FpcTP
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    solubility_product,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_constant import (
    ConstantKeq,
)


solver = get_solver("ipopt_v2")


def dummy_h(b, *args, **kwargs):
    return b.temperature


thermo_config = {
    # Specifying components
    "components": {
        "H2O": {"type": Component, "enth_mol_liq_comp": dummy_h},
        "Na+": {"type": Component, "enth_mol_liq_comp": dummy_h},
        "Cl-": {"type": Component, "enth_mol_liq_comp": dummy_h},
        "NaCl": {"type": Component, "enth_mol_sol_comp": dummy_h},
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Ideal,
            "component_list": ["H2O", "Na+", "Cl-"],
        },
        "Sol": {
            "type": SolidPhase,
            "equation_of_state": Ideal,
            "component_list": ["NaCl"],
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FpcTP,
    "state_bounds": {
        "flow_mol_phase_comp": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 450, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (1e5, pyunits.Pa),
    "temperature_ref": (300, pyunits.K),
}

# Solubility of NaCl in H2O is 360 g/L, or 6.16 mol/L
# 1L of H2O is 55.56 mol, results in Ksp = 8.235e-3 on mole frac basis
rxn_config = {
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "equilibrium_reactions": {
        "S1": {
            "stoichiometry": {
                ("Liq", "Na+"): 1,
                ("Liq", "Cl-"): 1,
                ("Sol", "NaCl"): -1,
            },
            "equilibrium_constant": ConstantKeq,
            "equilibrium_form": solubility_product,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {"k_eq_ref": (8.235e-3, None)},
        }
    },
}


class TestSingleState(object):
    @pytest.fixture
    def model(self):
        m = ConcreteModel()

        # Add a test thermo package for validation
        m.pparams = GenericParameterBlock(**thermo_config)
        m.rparams = GenericReactionParameterBlock(
            property_package=m.pparams, **rxn_config
        )

        m.state = m.pparams.build_state_block([0], defined_state=False)

        m.rxn = m.rparams.build_reaction_block(
            [0], state_block=m.state, has_equilibrium=True
        )
        m.rxn[0].eps = 1e-14

        return m

    @pytest.mark.component
    def test_saturated(self, model):
        assert model.state[0].phase_component_set == [
            ("Liq", "H2O"),
            ("Liq", "Na+"),
            ("Liq", "Cl-"),
            ("Sol", "NaCl"),
        ]

        model.state[0].temperature.fix(298.15)
        model.state[0].pressure.fix(101325)

        # Solve for saturated states (i.e. solubility product applied) by
        # fixing the flowrate of solids to a positive value.
        # Then start with a large amount of Na+ and solve for the flowrate of
        # Cl- which satisfies the solubility product (ignore electroneutrality)
        model.state[0].flow_mol_phase_comp["Liq", "H2O"].fix(55.56)
        model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(20)
        model.state[0].flow_mol_phase_comp["Liq", "Cl-"].set_value(2.5)
        model.state[0].flow_mol_phase_comp["Sol", "NaCl"].fix(1)

        for i in range(20, 2, -1):
            model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(i)

            results = solver.solve(model)

            assert check_optimal_termination(results)

            assert pytest.approx(8.235e-3, abs=1e-8) == value(
                model.state[0].mole_frac_phase_comp["Liq", "Na+"]
                * model.state[0].mole_frac_phase_comp["Liq", "Cl-"]
            )

    @pytest.mark.component
    def test_subsaturated(self, model):
        assert model.state[0].phase_component_set == [
            ("Liq", "H2O"),
            ("Liq", "Na+"),
            ("Liq", "Cl-"),
            ("Sol", "NaCl"),
        ]

        model.state[0].temperature.fix(298.15)
        model.state[0].pressure.fix(101325)

        # For subsaturated systems, fix the Na and Cl flows and check that no
        # solid is formed.
        model.state[0].flow_mol_phase_comp["Liq", "H2O"].fix(55.56)

        for i in range(11):
            for j in range(11):
                if i * j < 40:
                    # Zeroes result in evaluation errors, so use a small number
                    if i == 0:
                        i = 1e-8
                    if j == 0:
                        j = 1e-8

                    # reinitialize after each iteration
                    model.state[0].mole_frac_phase_comp["Liq", "H2O"].set_value(0.25)
                    model.state[0].mole_frac_phase_comp["Sol", "NaCl"].set_value(0.25)
                    model.state[0].mole_frac_phase_comp["Liq", "Na+"].set_value(0.25)
                    model.state[0].mole_frac_phase_comp["Liq", "Cl-"].set_value(0.25)

                    model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(i)
                    model.state[0].flow_mol_phase_comp["Liq", "Cl-"].fix(j)
                    model.state[0].flow_mol_phase_comp["Sol", "NaCl"].set_value(0)

                    iscale.constraint_scaling_transform(
                        model.rxn[0].equilibrium_constraint["S1"], 10
                    )

                    results = solver.solve(model)

                    assert check_optimal_termination(results)

                    assert pytest.approx(0, abs=1e-5) == value(
                        model.state[0].flow_mol_phase_comp["Sol", "NaCl"]
                    )


class TestUnit(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Add a test thermo package for validation
        m.fs.pparams = GenericParameterBlock(**thermo_config)
        m.fs.rparams = GenericReactionParameterBlock(
            property_package=m.fs.pparams, **rxn_config
        )

        # Don't include energy balances, as the test doesn't have a proper
        # enthalpy model. Fix outlet T instead.
        m.fs.R101 = EquilibriumReactor(
            property_package=m.fs.pparams,
            reaction_package=m.fs.rparams,
            has_equilibrium_reactions=True,
            has_rate_reactions=False,
            energy_balance_type=EnergyBalanceType.none,
        )

        m.fs.rparams.reaction_S1.eps = 1e-14
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(55.56)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "Na+"].fix(1e-8)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "Cl-"].fix(1e-8)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(1e-8)
        m.fs.R101.inlet.temperature[0].fix(298.15)
        m.fs.R101.inlet.pressure[0].fix(101325)

        m.fs.R101.outlet.temperature[0].fix(298.15)

        return m

    @pytest.mark.integration
    def test_subsaturated(self, model):
        for i in range(0, 6):
            # Values of zero cause problems, use a small number instead
            if i == 0:
                i == 1e-8

            model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(i)
            iscale.constraint_scaling_transform(
                model.fs.R101.control_volume.reactions[0.0].equilibrium_constraint[
                    "S1"
                ],
                10,
            )

            results = solver.solve(model)

            assert check_optimal_termination(results)

            assert pytest.approx(0, abs=1.1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"]
            )
            assert pytest.approx(i, rel=1e-6, abs=1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"]
            )
            assert pytest.approx(i, rel=1e-6, abs=1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"]
            )

    @pytest.mark.integration
    def test_saturated(self, model):
        model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(6.16)

        results = solver.solve(model)

        assert check_optimal_termination(results)

        assert pytest.approx(0, abs=1.1e-3) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"]
        )
        assert pytest.approx(6.159876, rel=1e-5) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"]
        )
        assert pytest.approx(6.159876, rel=1e-5) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"]
        )

    @pytest.mark.integration
    def test_supersaturated(self, model):
        for i in range(7, 21):
            model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(i)

            results = solver.solve(model)

            assert check_optimal_termination(results)

            assert pytest.approx(i, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"]
                + model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"]
            )
            assert pytest.approx(6.159876, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"]
            )
            assert pytest.approx(6.159876, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"]
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Eucken correlation to calculate low pressure pure component gas thermal conductivity,
(therm_cond_phase_comp) as outlined in The Properties of Gases and Liquids, 5th Ed.,
Section 10-3-1. Not particularly accurate, with errors up to 30% reported, especially
for polar compounds. The component-specific parameter f_int_eucken allows the user to
adjust the correlation on a component-by-component basis.
"""
from pyomo.environ import Var, units as pyunits
from idaes.core.util.misc import set_param_from_config
from idaes.core.util.constants import Constants
from idaes.core.util.exceptions import ConfigurationError


class Eucken(object):
    """Class to implement Eucken correlation for gas-phase thermal conductivity"""

    class therm_cond_phase_comp(object):
        """Eucken correlation for thermal conductivity"""

        @staticmethod
        def build_parameters(cobj, p):
            """Builds dimensionless parameter f_int"""
            # Properties of Gases and Liquids 10-3-1: Three common values for f_int: f_int=1, which is Eucken
            # classic, f_int=1.32, which is "Modified Eucken", and f_int=1.15, which is a compromise suggested by
            # Stiel, L. I., and G. Thodos:AIChE J., 10: 26 (1964)
            if not hasattr(cobj, "f_int_eucken"):
                cobj.f_int_eucken = Var(
                    doc="Dimensionless factor in Eucken formula associated with internal degrees of freedom",
                    units=pyunits.dimensionless,
                )
                set_param_from_config(cobj, param="f_int_eucken")

        @staticmethod
        def return_expression(b, cobj, p, T):
            """Returns expression for therm_cond_phase_comp"""
            # Properties of Gases and Liquids, Eq. 10-3.2
            units = b.params.get_metadata().derived_units

            M = pyunits.convert(cobj.mw, pyunits.kg / pyunits.mol)
            R = pyunits.convert(Constants.gas_constant, units["heat_capacity_mole"])
            f_int = cobj.f_int_eucken
            try:
                cp_mol_ig_comp = cobj.config.cp_mol_ig_comp
            except AttributeError:
                raise ConfigurationError(
                    f"Cannot find method to calculate cp_mol_ig_comp for component "
                    f"{cobj.local_name}."
                )

            if not hasattr(b, "_visc_d_phase_comp"):
                b._make_visc_d_phase_comp()  # pylint: disable=protected-access

            if hasattr(cp_mol_ig_comp, "return_expression"):
                cp_func = cp_mol_ig_comp.return_expression
            else:
                cp_func = cp_mol_ig_comp.cp_mol_ig_comp.return_expression

            therm_cond = (
                b._visc_d_phase_comp[  # pylint: disable=protected-access
                    p, cobj.local_name
                ]
                / M
                * (f_int * cp_func(b, cobj, T) + (15 / 4 - 5 * f_int / 2) * R)
            )

            return pyunits.convert(therm_cond, units["thermal_conductivity"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Air separation phase equilibrium package using Peng-Robinson EoS.

Example property package using the Generic Property Package Framework.
This example shows how to set up a property package to do air separation
phase equilibrium in the generic framework using Peng-Robinson equation
along with methods drawn from the pre-built IDAES property libraries.

The example includes two dictionaries.

1. The dictionary named configuration contains parameters obtained from
The Properties of Gases and Liquids (1987) 4th edition and NIST.

2. The dictionary named configuration_Dowling_2015 contains parameters used in
A framework for efficient large scale equation-oriented flowsheet optimization
(2015) Dowling. The parameters are extracted from Properties of Gases and
Liquids (1977) 3rd edition for Antoine's vapor equation and acentric factors
and converted values from the Properties of Gases and Liquids (1977)
3rd edition to j.
"""

# Import Python libraries
import logging

# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
from idaes.models.properties.modular_properties.pure import RPP4
from idaes.models.properties.modular_properties.pure import NIST
from idaes.models.properties.modular_properties.pure import RPP3

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for a Peng-Robinson Oxygen-Argon-Nitrogen system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] NIST, https://webbook.nist.gov/
#     Retrieved 16th August, 2020
# [3] The Properties of Gases and Liquids (1987)
#     3rd edition, Chemical Engineering Series - Robert C. Reid
#     Cp parameters where converted to j in Dowling 2015
# [4] A framework for efficient large scale equation-oriented flowsheet optimization (2015)
#     Computers and Chemical Engineering - Alexander W. Dowling

configuration = {
    # Specifying components
    "components": {
        "nitrogen": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (28.0135e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (34e5, pyunits.Pa),  # [1]
                "temperature_crit": (126.2, pyunits.K),  # [1]
                "omega": 0.037,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (3.115e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (-1.357e-2, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (2.680e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-1.168e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    191.61,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (3.7362, None),  # [2]
                    "B": (264.651, pyunits.K),
                    "C": (-6.788, pyunits.K),
                },
            },
        },
        "argon": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (39.948e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.98e5, pyunits.Pa),  # [1]
                "temperature_crit": (150.86, pyunits.K),  # [1]
                "omega": 0.001,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (2.050e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (0.0, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0.0, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0.0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    154.8,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (3.29555, None),  # [2]
                    "B": (215.24, pyunits.K),
                    "C": (-22.233, pyunits.K),
                },
            },
        },
        "oxygen": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (31.999e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (50.43e5, pyunits.Pa),  # [1]
                "temperature_crit": (154.58, pyunits.K),  # [1]
                "omega": 0.025,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (2.811e1, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (-3.680e-6, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (1.746e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-1.065e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    205.152,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (3.85845, None),  # [2]
                    "B": (325.675, pyunits.K),
                    "C": (-5.667, pyunits.K),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (10, 300, 350, pyunits.K),
        "pressure": (5e4, 1e5, 1e7, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("nitrogen", "nitrogen"): 0.000,
            ("nitrogen", "argon"): -0.26e-2,
            ("nitrogen", "oxygen"): -0.119e-1,
            ("argon", "nitrogen"): -0.26e-2,
            ("argon", "argon"): 0.000,
            ("argon", "oxygen"): 0.104e-1,
            ("oxygen", "nitrogen"): -0.119e-1,
            ("oxygen", "argon"): 0.104e-1,
            ("oxygen", "oxygen"): 0.000,
        }
    },
}

configuration_Dowling_2015 = {
    # Specifying components
    "components": {
        "nitrogen": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP3,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (28.0135e-3, pyunits.kg / pyunits.mol),  # [3]
                "pressure_crit": (33.943875e5, pyunits.Pa),  # [4]
                "temperature_crit": (126.2, pyunits.K),  # [4]
                "omega": 0.04,  # [3]
                "cp_mol_ig_comp_coeff": {
                    "A": (3.112896e1, pyunits.J / pyunits.mol / pyunits.K),  # [3]
                    "B": (-1.356e-2, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (2.6878e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-1.167e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    191.61,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (14.9342, None),  # [3]
                    "B": (588.72, pyunits.K),
                    "C": (-6.60, pyunits.K),
                },
            },
        },
        "argon": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP3,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (39.948e-3, pyunits.kg / pyunits.mol),  # [3]
                "pressure_crit": (48.737325e5, pyunits.Pa),  # [4]
                "temperature_crit": (150.86, pyunits.K),  # [4]
                "omega": -0.004,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (2.0790296e1, pyunits.J / pyunits.mol / pyunits.K),  # [3]
                    "B": (-3.209e-05, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (5.163e-08, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0.0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    154.8,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (15.2330, None),  # [3]
                    "B": (700.51, pyunits.K),
                    "C": (-5.84, pyunits.K),
                },
            },
        },
        "oxygen": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP3,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (31.999e-3, pyunits.kg / pyunits.mol),  # [3]
                "pressure_crit": (50.45985e5, pyunits.Pa),  # [4]
                "temperature_crit": (154.58, pyunits.K),  # [4]
                "omega": 0.021,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (2.8087192e1, pyunits.J / pyunits.mol / pyunits.K),  # [3]
                    "B": (-3.678e-6, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (1.745e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-1.064e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    205.152,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (15.4075, None),  # [3]
                    "B": (734.55, pyunits.K),
                    "C": (-6.45, pyunits.K),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (10, 300, 350, pyunits.K),
        "pressure": (5e4, 1e5, 1e7, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("nitrogen", "nitrogen"): 0.000,
            ("nitrogen", "argon"): -0.26e-2,
            ("nitrogen", "oxygen"): -0.119e-1,
            ("argon", "nitrogen"): -0.26e-2,
            ("argon", "argon"): 0.000,
            ("argon", "oxygen"): 0.104e-1,
            ("oxygen", "nitrogen"): -0.119e-1,
            ("oxygen", "argon"): 0.104e-1,
            ("oxygen", "oxygen"): 0.000,
        }
    },
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

# Changing existing config block attributes
# pylint: disable=protected-access

from enum import Enum

import pyomo.environ as pyo
from pyomo.common.config import ConfigValue, In

from idaes.core import declare_process_block_class
from idaes.models_extra.power_generation.unit_models.balance import BalanceBlockData
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.models.properties.helmholtz.helmholtz as hltz
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError


_log = idaeslog.getLogger(__name__)


class ValveFunctionType(Enum):
    linear = 1
    quick_opening = 2
    equal_percentage = 3
    custom = 4


def _assert_properties(pb):
    """Assert that the properties parameter block conforms to the requirements"""
    try:
        assert isinstance(pb, hltz.HelmholtzParameterBlockData)
        assert pb.config.phase_presentation in {
            hltz.PhaseType.MIX,
            hltz.PhaseType.L,
            hltz.PhaseType.G,
        }
        assert pb.config.state_vars == hltz.StateVars.PH
    except AssertionError:
        _log.error(
            "helm.HelmValve requires a Helmholtz EOS with "
            "a single or mixed phase and pressure-enthalpy state vars."
        )
        raise


def _linear_callback(blk):
    @blk.Expression(blk.flowsheet().time)
    def valve_function(b, t):
        return b.valve_opening[t]


def _quick_open_callback(blk):
    @blk.Expression(blk.flowsheet().time)
    def valve_function(b, t):
        return pyo.sqrt(b.valve_opening[t])


def _equal_percentage_callback(blk):
    blk.alpha = pyo.Var(initialize=1, doc="Valve function parameter")
    blk.alpha.fix()

    @blk.Expression(blk.flowsheet().time)
    def valve_function(b, t):
        return b.alpha ** (b.valve_opening[t] - 1)


def _liquid_pressure_flow_rule(b, t):
    """
    For liquid F = Cv*sqrt(Pi**2 - Po**2)*f(x)
    """
    Po = b.control_volume.properties_out[t].pressure
    Pi = b.control_volume.properties_in[t].pressure
    F = b.control_volume.properties_in[t].flow_mol
    fun = b.valve_function[t]
    return F**2 == b.Cv**2 * (Pi - Po) * fun**2


def _vapor_pressure_flow_rule(b, t):
    """
    For vapor F = Cv*sqrt(Pi**2 - Po**2)*f(x)
    """
    Po = b.control_volume.properties_out[t].pressure
    Pi = b.control_volume.properties_in[t].pressure
    F = b.control_volume.properties_in[t].flow_mol
    fun = b.valve_function[t]
    return F**2 == b.Cv**2 * (Pi**2 - Po**2) * fun**2


@declare_process_block_class("HelmValve")
class HelmValveData(BalanceBlockData):
    """
    Basic adiabatic 0D valve model.  This inherits the balance block to get
    a lot of unit model boilerplate and the mass balance, enegy balance and
    pressure equations.  This model is intended to be used only with Helmholtz
    EOS property packages in mixed or single phase mode with P-H state vars.

    Since this inherits BalanceBlockData, and only operates in steady-state or
    pseudo-steady-state (for dynamic models) the following mass, energy and
    pressure equations are implicitly written.

    1) Mass Balance:
        0 = flow_mol_in[t] - flow_mol_out[t]
    2) Energy Balance:
        0 = (flow_mol[t]*h_mol[t])_in - (flow_mol[t]*h_mol[t])_out
    3) Pressure:
        0 = P_in[t] + deltaP[t] - P_out[t]
    """

    CONFIG = BalanceBlockData.CONFIG()
    # For dynamics assume pseudo-steady-state
    CONFIG.dynamic = False
    CONFIG.get("dynamic")._default = False
    CONFIG.get("dynamic")._domain = In([False])
    CONFIG.has_holdup = False
    CONFIG.get("has_holdup")._default = False
    CONFIG.get("has_holdup")._domain = In([False])
    # Rest of config to make this function like a turbine
    CONFIG.has_pressure_change = True
    CONFIG.get("has_pressure_change")._default = True
    CONFIG.get("has_pressure_change")._domain = In([True])
    CONFIG.has_work_transfer = False
    CONFIG.get("has_work_transfer")._default = False
    CONFIG.get("has_work_transfer")._domain = In([False])
    CONFIG.has_heat_transfer = False
    CONFIG.get("has_heat_transfer")._default = False
    CONFIG.get("has_heat_transfer")._domain = In([False])
    CONFIG.declare(
        "valve_function",
        ConfigValue(
            default=ValveFunctionType.linear,
            domain=In(ValveFunctionType),
            description="Valve function type, if custom provide an expression rule",
            doc="""The type of valve function, if custom provide an expression rule
with the valve_function_rule argument.
**default** - ValveFunctionType.linear
**Valid values** - {
ValveFunctionType.linear,
ValveFunctionType.quick_opening,
ValveFunctionType.equal_percentage,
ValveFunctionType.custom}""",
        ),
    )
    CONFIG.declare(
        "valve_function_callback",
        ConfigValue(
            default=None,
            description="This is a callback that adds a valve function.  The "
            "callback function takes the valve bock data argument.",
        ),
    )
    CONFIG.declare(
        "phase",
        ConfigValue(
            default="Vap",
            domain=In(("Vap", "Liq")),
            description='Expected phase of fluid in valve in {"Liq", "Vap"}',
        ),
    )

    def build(self):
        """
        Add model equations to the unit model.  This is called by a default block
        construnction rule when the unit model is created.
        """
        super().build()  # Basic unit model build/read config
        config = self.config  # shorter config pointer

        # The thermodynamic expression writer object, te, writes expressions
        # including external function calls to calculate thermodynamic quantities
        # from a set of state variables.
        _assert_properties(config.property_package)

        self.valve_opening = pyo.Var(
            self.flowsheet().time,
            initialize=1,
            doc="Fraction open for valve from 0 to 1",
        )
        self.Cv = pyo.Var(
            initialize=0.1,
            doc="Valve flow coefficient, for vapor " "[mol/s/Pa] for liquid [mol/s/Pa]",
            units=pyo.units.mol / pyo.units.s / pyo.units.Pa,
        )
        # self.Cv.fix()

        # set up the valve function rule.  I'm not sure these matter too much
        # for us, but the options are easy enough to provide.
        vfcb = self.config.valve_function_callback
        vfselect = self.config.valve_function
        if vfselect is not ValveFunctionType.custom and vfcb is not None:
            _log.warning(
                "A valve function callback was provided but the valve "
                "function type is not custom."
            )

        if vfselect == ValveFunctionType.linear:
            _linear_callback(self)
        elif vfselect == ValveFunctionType.quick_opening:
            _quick_open_callback(self)
        elif vfselect == ValveFunctionType.equal_percentage:
            _equal_percentage_callback(self)
        else:
            if vfcb is None:
                raise ConfigurationError("No custom valve function callback provided")
            vfcb(self)

        if self.config.phase == "Liq":
            rule = _liquid_pressure_flow_rule
        else:
            rule = _vapor_pressure_flow_rule

        self.pressure_flow_equation = pyo.Constraint(self.flowsheet().time, rule=rule)

    def _get_performance_contents(self, time_point=0):
        """This returns a dictionary of quntities to be used in IDAES unit model
        report generation routines.
        """
        pc = super()._get_performance_contents(time_point=time_point)
        return pc

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        calculate_cv=False,
        calculate_opening=False,
    ):
        """
        For simplicity this initialization requires you to set values for the
        efficiency, inlet, and one of pressure ratio, pressure change or outlet
        pressure.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        init_log.info("Steam valve initialization started")

        # Create solver
        opt = get_solver(solver, optarg)

        # Store original specification so initialization doesn't change the model
        # This will only restore the values of variables that were originally fixed
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)
        # Check for alternate pressure specs
        for t in self.flowsheet().time:
            if self.outlet.pressure[t].fixed:
                self.deltaP[t].fix(
                    pyo.value(self.outlet.pressure[t] - self.inlet.pressure[t])
                )
                self.outlet.pressure[t].unfix()
            elif self.deltaP[t].fixed:
                # No outlet pressure specified guess a small pressure drop
                self.outlet.pressure[t] = pyo.value(
                    self.inlet.pressure[t] + self.deltaP[t]
                )

        self.inlet.fix()
        self.outlet.unfix()
        for t, v in self.deltaP.items():
            if calculate_cv:
                self.Cv.unfix()
            elif calculate_opening:
                self.valve_opening.unfix()
            elif v.fixed and self.pressure_flow_equation.active:
                self.inlet.flow_mol[t].unfix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            opt.solve(self, tee=slc.tee)

        init_log.info("Steam valve initialization complete")

        from_json(self, sd=istate, wts=sp)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for t, c in self.pressure_flow_equation.items():
            s = iscale.get_scaling_factor(self.control_volume.properties_in[t].flow_mol)
            s = s**2
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Basic test to make sure the pynumero library is available and working
"""

import pyomo.environ as pyo
import pytest

try:
    from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
    from pyomo.contrib.pynumero.asl import AmplInterface
except ImportError:
    PyomoNLP = None
    AmplInterface = None


@pytest.fixture
def model():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3], initialize=4.0)
    m.c = pyo.Constraint(expr=m.x[3] ** 2 + m.x[1] == 25)
    m.d = pyo.Constraint(expr=m.x[2] ** 2 + m.x[1] <= 18.0)
    m.o = pyo.Objective(expr=m.x[1] ** 4 - 3 * m.x[1] * m.x[2] ** 3 + m.x[3] ** 2 - 8.0)
    m.x[1].setlb(0.0)
    m.x[2].setlb(0.0)
    return m


@pytest.mark.unit
def test_import():
    assert PyomoNLP is not None


@pytest.mark.unit
def test_have_pynumero(model):
    assert PyomoNLP is not None
    assert AmplInterface.available()
    nlp = PyomoNLP(model)
    f = nlp.evaluate_objective()
    assert f == pytest.approx(-504.0)

    jac = nlp.evaluate_jacobian().toarray()
    assert jac[0][0] == pytest.approx(0)
    assert jac[0][1] == pytest.approx(8)
    assert jac[0][2] == pytest.approx(1)
    assert jac[1][0] == pytest.approx(8)
    assert jac[1][1] == pytest.approx(0)
    assert jac[1][2] == pytest.approx(1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Flowsheet example of the MB model for a methane combustion with
iron-oxide case study


Author: Chinedu Okoli
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

import time

from pyomo.environ import ConcreteModel, value

from idaes.core import FlowsheetBlock
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

# Import IDAES logger
import idaes.logger as idaeslog

# Import MBR unit model
from idaes.models_extra.gas_solid_contactors.unit_models.moving_bed import MBR

# Import property packages
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.hetero_reactions import (
    HeteroReactionParameterBlock,
)


# -----------------------------------------------------------------------------
def main():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up thermo props and reaction props
    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.solid_properties = SolidPhaseParameterBlock()

    m.fs.hetero_reactions = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )

    m.fs.MB = MBR(
        transformation_method="dae.collocation",
        gas_phase_config={"property_package": m.fs.gas_properties},
        solid_phase_config={
            "property_package": m.fs.solid_properties,
            "reaction_package": m.fs.hetero_reactions,
        },
    )

    # Fix bed geometry variables
    m.fs.MB.bed_diameter.fix(6.5)  # m
    m.fs.MB.bed_height.fix(5)  # m

    # Fix inlet port variables for gas and solid
    m.fs.MB.gas_inlet.flow_mol[0].fix(128.20513)  # mol/s
    m.fs.MB.gas_inlet.temperature[0].fix(298.15)  # K
    m.fs.MB.gas_inlet.pressure[0].fix(2.00e5)  # Pa = 1E5 bar
    m.fs.MB.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.02499)
    m.fs.MB.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.00001)
    m.fs.MB.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.975)

    m.fs.MB.solid_inlet.flow_mass[0].fix(591.4)  # kg/s
    # Particle porosity:
    # The porosity of the OC particle at the inlet is calculated from the
    # known bulk density of the fresh OC particle (3251.75 kg/m3), and the
    # skeletal density of the fresh OC particle (calculated from the known
    # composition of the fresh particle, and the skeletal density of its
    # components [see the solids property package])
    m.fs.MB.solid_inlet.particle_porosity[0].fix(0.27)
    m.fs.MB.solid_inlet.temperature[0].fix(1183.15)  # K
    m.fs.MB.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
    m.fs.MB.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
    m.fs.MB.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

    # Initialize fuel reactor
    t_start = time.time()  # Run start time

    # State arguments for initializing property state blocks
    # Gas phase temperature is initialized at solid
    # temperature because thermal mass of solid >> thermal mass of gas
    # Particularly useful for initialization if reaction takes place
    blk = m.fs.MB
    gas_phase_state_args = {
        "flow_mol": blk.gas_inlet.flow_mol[0].value,
        "temperature": blk.solid_inlet.temperature[0].value,
        "pressure": blk.gas_inlet.pressure[0].value,
        "mole_frac": {
            "CH4": blk.gas_inlet.mole_frac_comp[0, "CH4"].value,
            "CO2": blk.gas_inlet.mole_frac_comp[0, "CO2"].value,
            "H2O": blk.gas_inlet.mole_frac_comp[0, "H2O"].value,
        },
    }
    solid_phase_state_args = {
        "flow_mass": blk.solid_inlet.flow_mass[0].value,
        "particle_porosity": blk.solid_inlet.particle_porosity[0].value,
        "temperature": blk.solid_inlet.temperature[0].value,
        "mass_frac": {
            "Fe2O3": blk.solid_inlet.mass_frac_comp[0, "Fe2O3"].value,
            "Fe3O4": blk.solid_inlet.mass_frac_comp[0, "Fe3O4"].value,
            "Al2O3": blk.solid_inlet.mass_frac_comp[0, "Al2O3"].value,
        },
    }

    print()
    print("Apply scaling transformation")
    # Scale the model by applying scaling transformation
    # This reduces ill conditioning of the model
    iscale.calculate_scaling_factors(m)

    print()
    print("Initialize the model")
    m.fs.MB.initialize(
        outlvl=idaeslog.INFO,
        optarg={"tol": 1e-5},
        gas_phase_state_args=gas_phase_state_args,
        solid_phase_state_args=solid_phase_state_args,
    )

    t_initialize = time.time()  # Initialization time

    print()
    print("Solve the model")
    # Create a solver
    solver = get_solver()
    solver.solve(m.fs.MB, tee=True, options={"tol": 1e-5})

    t_simulation = time.time()  # Simulation time

    print("\n")
    print("----------------------------------------------------------")
    print("Total initialization time: ", value(t_initialize - t_start), " s")
    print("----------------------------------------------------------")

    print("\n")
    print("----------------------------------------------------------")
    print("Total simulation time: ", value(t_simulation - t_start), " s")
    print("----------------------------------------------------------")

    return m


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    m = main()
    # TODO: Fix this
    # pylint: disable-next=protected-access
    stream_table = m.fs.MB._get_stream_table_contents()
    print(stream_table)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Initializer class for implementing initialization from a data source
"""
from idaes.core.initialization.initializer_base import InitializerBase

__author__ = "Andrew Lee"


class FromDataInitializer(InitializerBase):
    """
    This is a general purpose Initializer object which attempts to initialize a
    model from user provided data.

    Data can be provided in either json format or as a dict-like structure. The loaded
    solution is then checked to ensure that it satisfies all constraints in the model.

    """

    def initialization_routine(self, model):
        # No action required as data has been loaded in the load_initial_guesses method
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from pyomo.core.expr.numeric_expr import MonomialTermExpression
from idaes.apps.matopt.opt.pyomo_modeling import getLB, getUB
from pyomo.environ import *
import pytest


@pytest.mark.unit
def test_getLB():
    mod = ConcreteModel()
    mod.x = Var(within=NonNegativeReals, bounds=(0, 2), initialize=1)
    mod.e = MonomialTermExpression((1, mod.x))
    assert isinstance(mod.e, MonomialTermExpression)
    LB = getLB(mod.e)


@pytest.mark.unit
def test_getUB():
    mod = ConcreteModel()
    mod.x = Var(within=NonNegativeReals, bounds=(0, 2), initialize=1)
    mod.e = MonomialTermExpression((1, mod.x))
    assert isinstance(mod.e, MonomialTermExpression)
    UB = getUB(mod.e)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring


def assert_disjoint_intervals(intervals):
    """
    This function takes intervals in the form of tuples and makes sure
    that they are disjoint.

    Arguments
    ---------
    intervals: iterable
        Iterable of tuples, each containing the low and high values of an
        interval.

    """
    intervals = list(sorted(intervals))
    for i, (lo, hi) in enumerate(intervals):
        if not lo <= hi:
            raise RuntimeError(
                "Lower endpoint of interval is higher than upper endpoint"
            )
        if i != 0:
            prev_lo, prev_hi = intervals[i - 1]
            if not prev_hi <= lo:
                raise RuntimeError(
                    "Intervals %s and %s are not disjoint"
                    % ((prev_lo, prev_hi), (lo, hi))
                )


def load_inputs_into_model(model, time, input_data, time_tol=0):
    """
    This function loads piecewise constant values into variables (or
    mutable parameters) of a model.

    Arguments
    ---------
    model: _BlockData
        Pyomo block containing the variables and parameters whose values
        will be set
    time: ContinuousSet
        Pyomo ContinuousSet corresponding to the piecewise constant intervals
    input_data: dict of dicts
        Maps variable names to dictionaries mapping 2-tuples to values.
        Each tuple contains the low and high endpoints of an interval
        on which the variable is to take the specified value.
    time_tol: float
        Optional. Tolerance within which the ContinuousSet will be searched
        for interval endpoints. The default is zero, i.e. the endpoints
        must be within the ContinuousSet exactly.

    """
    for cuid, inputs in input_data.items():
        var = model.find_component(cuid)
        if var is None:
            raise RuntimeError(
                "Could not find a variable on model %s with ComponentUID %s"
                % (model.name, cuid)
            )

        intervals = list(sorted(inputs.keys()))
        assert_disjoint_intervals(intervals)
        for i, interval in enumerate(intervals):  # pylint: disable=unused-variable
            idx0 = time.find_nearest_index(interval[0], tolerance=time_tol)
            idx1 = time.find_nearest_index(interval[1], tolerance=time_tol)
            if idx0 is None or idx1 is None:
                # One of the interval boundaries is not a valid time index
                # within tolerance. Skip this interval and move on.
                continue
            input_val = inputs[interval]
            idx_iter = range(idx0 + 1, idx1 + 1) if idx0 != idx1 else (idx0,)
            for idx in idx_iter:
                t = time.at(idx)
                var[t].set_value(input_val)


def interval_data_from_time_series(data, use_left_endpoint=False):
    """
    This function converts time series data to piecewise constant
    interval data. A series of N time points and values yields
    N-1 intervals. By default, each interval takes the value of
    its right endpoint.

    In:
        ([t0, ...], {str(cuid): [value0, ...], },)
    Out:
        {str(cuid): {(t0, t1): value0 or value1, ...},}

    Arguments
    ---------
    data: tuple
        First entry is a list of time points, second entry is a dict
        mapping names each to a list of values at the corresponding time
        point
    use_left_endpoint: bool
        Optional. Indicates whether each interval should take the value
        of its left endpoint. Default is False, i.e. each interval takes
        the value of its right endpoint.

    Returns
    -------
    dict of dicts
        Maps names to a dict that maps interval-tuples each to the value
        over that interval

    """
    time, value_dict = data
    n_t = len(time)
    if n_t == 1:
        t0 = time[0]
        return {name: {(t0, t0): values[0]} for name, values in value_dict.items()}
    else:
        # This covers the case of n_t > 1 and n_t == 0
        interval_data = {}
        intervals = [(time[i - 1], time[i]) for i in range(1, n_t)]
        for name, values in value_dict.items():
            interval_values = [
                values[i - 1] if use_left_endpoint else values[i] for i in range(1, n_t)
            ]
            interval_data[name] = dict(zip(intervals, interval_values))
        return interval_data



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pyomo.environ as pyo

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
import idaes.core.util.scaling as iscale
from idaes.models_extra.power_generation.properties.natural_gas_PR import get_prop
from idaes.models_extra.power_generation.unit_models.soec_design import (
    SoecDesign,
    EosType,
)
from idaes.core.solvers import get_solver
import pytest


def flowsheet(eos=EosType.PR):
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    sweep_comp = {
        "O2": 0.2074,
        "H2O": 0.0099,
        "CO2": 0.0003,
        "N2": 0.7732,
        "Ar": 0.0092,
    }

    feed_comp = {
        "H2": 0.01,
        "H2O": 0.99,
    }

    m.fs.o2_side_prop_params = GenericParameterBlock(
        **get_prop(sweep_comp, {"Vap"}, eos=eos), doc="Air property parameters"
    )
    m.fs.h2_side_prop_params = GenericParameterBlock(
        **get_prop(feed_comp, {"Vap"}, eos=eos), doc="Flue gas property parameters"
    )
    m.fs.soec = SoecDesign(
        oxygen_side_property_package=m.fs.o2_side_prop_params,
        hydrogen_side_property_package=m.fs.h2_side_prop_params,
        reaction_eos=eos,
    )

    m.fs.soec.hydrogen_side_inlet.temperature.fix(1023)
    m.fs.soec.hydrogen_side_inlet.pressure.fix(20e5)
    m.fs.soec.hydrogen_side_inlet.flow_mol.fix(2)
    for c, v in feed_comp.items():
        m.fs.soec.hydrogen_side_inlet.mole_frac_comp[:, c].fix(v)
    m.fs.soec.oxygen_side_inlet.temperature.fix(1023)
    m.fs.soec.oxygen_side_inlet.pressure.fix(20e5)
    m.fs.soec.oxygen_side_inlet.flow_mol.fix(2)
    for c, v in sweep_comp.items():
        m.fs.soec.oxygen_side_inlet.mole_frac_comp[:, c].fix(v)
    m.fs.soec.hydrogen_side_outlet_temperature.fix(1023)
    m.fs.soec.oxygen_side_outlet_temperature.fix(1023)
    m.fs.soec.water_utilization.fix(0.7)
    iscale.calculate_scaling_factors(m)
    m.fs.soec.initialize(optarg={"max_iter": 30})
    return m


@pytest.mark.component
def test_soec_design_ideal():
    m = flowsheet(eos=EosType.IDEAL)
    solver = get_solver(solver="ipopt")
    res = solver.solve(m)
    # Make sure it converged
    assert pyo.check_optimal_termination(res)
    # Based on the inlet composition and the water utilization I know this
    assert pytest.approx(0.703) == pyo.value(
        m.fs.soec.hydrogen_side_outlet.mole_frac_comp[0, "H2"]
    )
    # should be at the thermoneutral voltage and I know about what that is
    assert pytest.approx(1.287, abs=0.02) == pyo.value(m.fs.soec.cell_potential[0])


@pytest.mark.component
def test_soec_design_pr():
    m = flowsheet(eos=EosType.IDEAL)
    solver = get_solver(solver="ipopt")
    res = solver.solve(m)
    # Make sure it converged
    assert pyo.check_optimal_termination(res)
    # Based on the inlet composition and the water utilization I know this
    assert pytest.approx(0.703) == pyo.value(
        m.fs.soec.hydrogen_side_outlet.mole_frac_comp[0, "H2"]
    )
    # should be at the thermoneutral voltage and I know about what that is
    assert pytest.approx(1.287, abs=0.02) == pyo.value(m.fs.soec.cell_potential[0])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
The basic functionality is to allow a user to add any model of choice that
can interface with other IDAES unit models that leverage the ControlVolume
blocks. This is done by letting the user create ports and populate with
members that match with state variables defined in an IDAES unit model.
Also, a default initialize method is provided with a flexibility to pass a
callback from the user.
"""

from pyomo.network import Port
from pyomo.common.config import ConfigValue, In

from idaes.core.base.process_base import declare_process_block_class, ProcessBlockData
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
from idaes.core.initialization import BlockTriangularizationInitializer

__author__ = "Jaffer Ghouse"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SkeletonUnitModel")
class SkeletonUnitModelData(ProcessBlockData):
    """
    This is the class for a skeleton unit model. This will provide a user
    with a generic skeleton to add a custom or a surrogate unit model
    when control volumes are not required.
    """

    # Set default initializer
    # TODO: For now, use Block Triangularization as the default
    default_initializer = BlockTriangularizationInitializer

    # This is a staticmethod that will be the default callable set for the
    # initializer flag in the config block.
    @staticmethod
    def _default_initializer(
        model, opt=None, init_log=None, solve_log=None, initial_guess=None
    ):
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(model, tee=slc.tee)
        init_log.info(
            "Initialization completed using default method {}.".format(
                idaeslog.condition(res)
            )
        )

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Dynamic model flag",
            doc="""Model does not support dynamics or holdup automatically.
        The variables declared in this unit will need to be indexed with the
        flowsheet time domain to facilitate connection with other
        unit models.""",
        ),
    )
    CONFIG.declare(
        "initializer",
        ConfigValue(
            # Note: staticmethod is unbound, extract with __func__
            default=_default_initializer.__func__,
            description="Initializer to be used",
            doc="""Flag to set the callback from user for initialization.
        Default is set to use the default_initialize method.""",
        ),
    )

    # pylint: disable=W0235
    def build(self):
        """
        General build method for SkeletonUnitModelBlockData.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super().build()

    def add_ports(self, name, member_dict, doc=None):
        """
        This is a method to build Port objects in the SkeletonUnitModel and
        populate them with appropriate port members as specified. User can add
        as many inlet and outlet ports as required.

        Keyword Args:
            name : name to use for Port object.
            dict : dictionary containing variables to be added to the port
            doc : doc string for Port object

        Returns:
            A Pyomo Port object and associated components.
        """
        # Validate that member_dict is a dict
        if not isinstance(member_dict, dict):
            raise ConfigurationError(
                "member_dict should be a dictionary "
                "with the keys being the name assigned(strings) and "
                "values being the variable objects declared in the model."
            )

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)

        # Add port object to model
        setattr(self, name, p)

        # Populate port and map names to actual variables as defined
        for k in member_dict.keys():
            p.add(member_dict[k], name=k)

    def fix_initialization_states(self):
        """
        Attempts to fix inlet states by iterating over all Ports and looking for "inlet"
        in the name.

        Users may need to fix inlets manually if they use names which do not include "inlet".

        Returns:
            None
        """
        for p in self.component_objects(Port, descend_into=False):
            if "inlet" in p.name:
                # Assume name is correct and fix all variables inside
                p.fix()

    # TODO : Work out how to make this work with new UnitModel initialization
    def initialize(
        self, outlvl=idaeslog.NOTSET, solver=None, optarg=None, initial_guess=None
    ):
        """Initialize method for the SkeletonUnitModel. If a custom function
        is provided via the initializer argument in the config block,
        then, this method will use it. If no custom function is
        provided, the _default_initializer method is used. Irrespective
        of which method is being used (default or custom), the expectation is
        that the degrees of freedom is zero at the start of initialization.

        Args:
            outlvl (idaes logger, optional):
                Set idaes logger level. Defaults to idaeslog.NOTSET.
            solver (string, optional):
                solver to be used for solve. Defaults to None but if None,
                ipopt from IDAES is used.
            optarg ({dict}, optional): solver arguments. Defaults to None
                but if None, default args for ipopt from IDAES is used.
            initial_guess ({dict}, optional): initial guess that is passed
            to custom initialize. Not used for the default initialize method.

        Raises:
            ConfigurationError: If degrees of freedom is not zero at the start
            of initialization.
        """

        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver=solver, options=optarg)

        if degrees_of_freedom(self) == 0:
            self.config.initializer(
                self,
                opt=opt,
                init_log=init_log,
                solve_log=solve_log,
                initial_guess=initial_guess,
            )
        else:
            raise ConfigurationError(
                "Degrees of freedom is not zero during start of "
                "initialization. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generate parameter and expression files for propane
"""

__author__ = "John Eslick"

import os
import math
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from idaes.models.properties.general_helmholtz.helmholtz_parameters import (
    WriteParameters,
)
from idaes.core.util.math import smooth_max
from idaes.core.util.constants import Constants


def thermal_conductivity_rule(m):
    """
    Thermal conductivity expression rule.  See:

    Marsh, K., Perkins, R., and Ramires, M.L.V., "Measurement and Correlation
        of the Thermal Conductivity of Propane from 86 to 600 K at Pressures
        to 70 MPa, J. Chem. Eng. Data, 47(4):932-940, 2002.
    """
    a = [
        -0.00124778,
        0.00816371,
        0.0199374,
    ]
    b = {
        (1, 1): -0.03695,
        (1, 2): 0.0482798,
        (2, 1): 0.148658,
        (2, 2): -0.135636,
        (3, 1): -0.119986,
        (3, 2): 0.117588,
        (4, 1): 0.0412431,
        (4, 2): -0.0436911,
        (5, 1): -0.00486905,
        (5, 2): 0.00616079,
    }

    T = m.T_star / m.tau
    MW = m.MW / 1000.0  # kg/mol
    rho_star = m.rho_star / MW
    rho = m.delta * rho_star  # mol/m^3
    tau = m.tau
    delta = m.delta
    rho_crit = 5000  # mol/m^3
    # lamb_red = 1  # reducing tc W/m/K
    Tref = 554.73  # reference T K
    k = Constants.boltzmann_constant
    Pc = 4.2512  # MPa
    big_gam = 0.0496
    R0 = 1.03
    gamma = 1.239
    qd = 1 / 7.16635e-10
    xi0 = 1.94e-10  # m
    nu = 0.63
    Tred = 1.0 / m.tau
    l0 = sum(ai * Tred**i for i, ai in enumerate(a))
    lr = sum(m.delta**i * (b[i, 1] + b[i, 2] * Tred) for i in range(1, 6))

    m.cp = pyo.ExternalFunction(library="", function="cp")
    m.cv = pyo.ExternalFunction(library="", function="cv")
    m.mu = pyo.ExternalFunction(library="", function="mu")
    m.itc = pyo.ExternalFunction(library="", function="itc")

    drho_dp = m.itc("propane", delta, m.T_star / T) * rho_star * m.delta
    drho_dp_ref = m.itc("propane", delta, m.T_star / Tref) * rho_star * m.delta

    deltchi = smooth_max(
        Pc * rho / rho_crit**2 * (drho_dp - drho_dp_ref * Tref / T), 0, eps=1e-8
    )
    xi = xi0 * (deltchi / big_gam) ** (nu / gamma)

    cp = m.cp("propane", delta, tau)
    cv = m.cv("propane", delta, tau)
    mu = m.mu("propane", delta, tau) / 1e6
    y = qd * xi
    kappa_inv = cv / cp
    pi = math.pi
    Omega = 2.0 / pi * ((1.0 - kappa_inv) * pyo.atan(y) + kappa_inv * y)
    Omega0 = (
        2.0
        / pi
        * (1.0 - pyo.exp(-1.0 / (1.0 / y + 1.0 / 3.0 * (y * rho_crit / rho) ** 2)))
    )
    lambda_c = (
        1000 * MW * cp * rho * R0 * k * T / 6.0 / math.pi / mu / xi * (Omega - Omega0)
    )
    return 1000 * (l0 + lr + lambda_c)


def viscosity_rule(m):
    """
    Viscosity expression rule.  See:

    Vogel E, C Küchenmeister, E Bich, A Laesecke, Reference Correlation of the
        Viscosity of Propane. Journal of Physical and Chemical Reference Data
        27, 947–970 (1998)
    """
    a = {
        0: 0.25104574,
        1: -0.47271238,
        3: 0.060836515,
    }
    b = {
        0: -19.572881,
        1: 219.73999,
        2: -1015.3226,
        3: 2471.01251,
        4: -3375.1717,
        5: 2491.6597,
        6: -787.26086,
        7: 14.085455,
        8: -0.34664158,
    }
    t = {
        0: 0.0,
        1: -0.25,
        2: -0.50,
        3: -0.75,
        4: -1.00,
        5: -1.25,
        6: -1.50,
        7: -2.50,
        8: -5.50,
    }
    e = {
        (2, 0): 35.9873030195,
        (2, 1): -180.512188564,
        (2, 2): 87.7124888223,
        (3, 0): -105.773052525,
        (3, 1): 205.319740877,
        (3, 2): -129.210932610,
        (4, 0): 58.9491587759,
        (4, 1): -129.740033100,
        (4, 2): 76.6280419971,
        (5, 0): -9.59407868475,
        (5, 1): 21.0726986598,
        (5, 2): -14.3971968187,
    }
    f = {1: 1616.88405374}
    g = {1: 2.50053938863, 2: 0.860516059264}

    M = m.MW  # kmol/kg
    T = m.T_star / m.tau  # K
    tau = T / m.Tc
    rho = m.delta * m.rho_star / M  # mol/l
    sigma = 0.49748  # nm
    eok = 263.88  # K
    NA = 6.0221408e23  # number/mol
    Ts = T / eok  # dimensionless
    vs = pyo.exp(sum(ai * pyo.log(Ts) ** i for i, ai in a.items()))
    etas = 0.021357 * pyo.sqrt(M * T) / (sigma**2 * vs)

    Bs = sum(bi * Ts ** t[i] for i, bi in b.items())
    B = NA * sigma**3 * Bs * 1e-24  # l/mol
    delta0 = g[1] * (1 + g[2] * pyo.sqrt(tau))
    delta = rho / 5.0
    eta = sum(e[i, j] * delta**i / tau**j for i, j in e) + f[1] * (
        delta / (delta0 - delta) - delta / delta0
    )
    return etas * (1 + B * rho) + eta


def main(dry_run=False):
    """Generate parameter and expression files.

    Args:
        dry_run (bool): If dry run don't generate files

    Returns:
        None
    """
    main_param_file = os.path.join(this_file_dir(), "propane.json")
    we = WriteParameters(parameters=main_param_file)
    we.add(
        {
            "viscosity": viscosity_rule,
            "thermal_conductivity": thermal_conductivity_rule,
        }
    )
    we.write(dry_run=dry_run)
    return we


if __name__ == "__main__":
    main()



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import logging
import pytest

from io import StringIO
from pyomo.common.log import LoggingIntercept
from idaes.beta import import_beta


@pytest.mark.unit
def test_beta_module_exception():
    with pytest.raises(
        ImportError,
        match=r"Module 'idaes.tests.beta_mod' is in beta "
        "and must be imported using idaes.beta.import_beta().",
    ):
        import idaes.tests.beta_mod


@pytest.mark.unit
def test_beta_module_import():
    os = StringIO()
    with LoggingIntercept(os, "idaes", logging.INFO):
        mod = import_beta("idaes.tests.beta_mod")
    assert mod.__name__ == "idaes.tests.beta_mod"
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    os = StringIO()
    with LoggingIntercept(os, "idaes", logging.INFO):
        mod = import_beta(".beta_mod")
    assert mod.__name__ == "idaes.tests.beta_mod"
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    os = StringIO()
    with LoggingIntercept(os, "idaes.logger", logging.INFO):
        mod = import_beta("idaes.logger")
    assert mod.__name__ == "idaes.logger"
    assert (
        os.getvalue().strip() == "Module 'idaes.tests.test_beta' "
        "imported module 'idaes.logger' as a Beta module. "
        "This module is not declared beta and can be "
        "imported using Python's normal import mechanisms."
    )


@pytest.mark.unit
def test_beta_reimport():
    os = StringIO()
    with LoggingIntercept(os, "idaes", logging.INFO):
        mod = import_beta("idaes.tests.beta_mod")
    assert mod.__name__ == "idaes.tests.beta_mod"
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    # Test that subsequent standard imports will still trigger an
    # exception, even if the module was previously imported using
    # import-beta()
    with pytest.raises(
        ImportError,
        match=r"Module 'idaes.tests.beta_mod' is in beta "
        "and must be imported using idaes.beta.import_beta().",
    ):
        import idaes.tests.beta_mod

    # But subsequent beta_import() modules return the same (original)
    # module object
    os = StringIO()
    with LoggingIntercept(os, "idaes", logging.INFO):
        mod2 = import_beta("idaes.tests.beta_mod")
    assert mod2 is mod



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains classes for property blocks and property parameter blocks.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

import sys

# Import Pyomo libraries
from pyomo.environ import Set, value, Var, Expression, Constraint, Reference
from pyomo.core.base.var import VarData
from pyomo.core.base.expression import ExpressionData
from pyomo.common.config import ConfigBlock, ConfigValue, Bool
from pyomo.common.formatting import tabular_writer
from pyomo.network import Port

# Import IDAES cores
from idaes.core.base.process_block import ProcessBlock
from idaes.core import ProcessBlockData, MaterialFlowBasis
from idaes.core.base import property_meta
from idaes.core.base.phases import PhaseData
from idaes.core.base.components import ComponentData
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import (
    ConfigurationError,
    PropertyPackageError,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_activated_constraints,
    number_activated_blocks,
)
from idaes.core.util import scaling as iscale
import idaes.logger as idaeslog
from idaes.core.base.util import build_on_demand
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
)


# Some more information about this module
__author__ = "Andrew Lee, John Eslick"

__all__ = ["StateBlockData", "StateBlock", "PhysicalParameterBlock"]

# Set up logger
_log = idaeslog.getLogger(__name__)


class _lock_attribute_creation_context(object):
    """Context manager to lock creation of new attributes on a state block"""

    def __init__(self, block):
        self.block = block

    def __enter__(self):
        self.block._lock_attribute_creation = True

    def __exit__(self, exc_type, exc_value, traceback):
        self.block._lock_attribute_creation = False


class PhysicalParameterBlock(ProcessBlockData, property_meta.HasPropertyClassMetadata):
    """
    This is the base class for thermophysical parameter blocks. These are
    blocks that contain a set of parameters associated with a specific
    thermophysical property package, and are linked to by all instances of
    that property package.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "default_arguments",
        ConfigBlock(
            implicit=True, description="Default arguments to use with Property Package"
        ),
    )

    def build(self):
        """
        General build method for PropertyParameterBlocks. Inheriting models
        should call super().build.

        Args:
            None

        Returns:
            None
        """
        super(PhysicalParameterBlock, self).build()

        # Need this to work with the Helmholtz EoS package
        if not hasattr(self, "_state_block_class"):
            self._state_block_class = None

        # By default, property packages do not include inherent reactions
        self._has_inherent_reactions = False

    @property
    def state_block_class(self):
        if self._state_block_class is not None:
            return self._state_block_class
        else:
            raise AttributeError(
                "{} has not assigned a StateBlock class to be associated "
                "with this property package. Please contact the developer of "
                "the property package.".format(self.name)
            )

    @property
    def has_inherent_reactions(self):
        return self._has_inherent_reactions

    def build_state_block(self, *args, **kwargs):
        """
        Methods to construct a StateBlock associated with this
        PhysicalParameterBlock. This will automatically set the parameters
        construction argument for the StateBlock.

        Returns:
            StateBlock

        """
        # default = kwargs.pop("default", {})
        initialize = kwargs.pop("initialize", {})

        # TODO: Should find a better way to do this
        # For now, trigger get_phase_component_set to make sure
        # it has been constructed before building the state block
        self.get_phase_component_set()

        if initialize == {}:
            kwargs["parameters"] = self
        else:
            for i in initialize.keys():
                initialize[i]["parameters"] = self

        return self.state_block_class(  # pylint: disable=not-callable
            *args, **kwargs, initialize=initialize
        )

    def get_phase_component_set(self):
        """
        Method to get phase-component set for property package. If a phase-
        component set has not been constructed yet, this method will construct
        one.

        Args:
            None

        Returns:
            Phase-Component Set object
        """
        try:
            return self._phase_component_set
        except AttributeError:
            # Phase-component set does not exist, so create one.
            pc_set = []
            for p in self.phase_list:
                p_obj = self.get_phase(p)
                if p_obj.config.component_list is not None:
                    c_list = p_obj.config.component_list
                else:
                    c_list = self.component_list
                for j in c_list:
                    pc_set.append((p, j))

            self._phase_component_set = Set(initialize=pc_set, ordered=True)

            return self._phase_component_set

    def get_component(self, comp):
        """
        Method to retrieve a Component object based on a name from the
        component_list.

        Args:
            comp: name of Component object to retrieve

        Returns:
            Component object
        """
        obj = getattr(self, comp)
        if not isinstance(obj, ComponentData):
            raise PropertyPackageError(
                "{} get_component found an attribute {}, but it does not "
                "appear to be an instance of a Component object.".format(
                    self.name, comp
                )
            )
        return obj

    def get_phase(self, phase):
        """
        Method to retrieve a Phase object based on a name from the phase_list.

        Args:
            phase: name of Phase object to retrieve

        Returns:
            Phase object
        """
        obj = getattr(self, phase)
        if not isinstance(obj, PhaseData):
            raise PropertyPackageError(
                "{} get_phase found an attribute {}, but it does not "
                "appear to be an instance of a Phase object.".format(self.name, phase)
            )
        return obj


class StateBlock(ProcessBlock):
    """
    This is the base class for state block objects. These are used when
    constructing the SimpleBlock or IndexedBlock which will contain the
    PropertyData objects, and contains methods that can be applied to
    multiple StateBlockData objects simultaneously.
    """

    # Set default initializer
    default_initializer = BlockTriangularizationInitializer

    @property
    def component_list(self):
        return self._return_component_list()

    def _return_component_list(self):
        return self._get_parameter_block().component_list

    @property
    def phase_list(self):
        return self._return_phase_list()

    def _return_phase_list(self):
        return self._get_parameter_block().phase_list

    @property
    def phase_component_set(self):
        return self._return_phase_component_set()

    def _return_phase_component_set(self):
        return self._get_parameter_block().get_phase_component_set()

    @property
    def has_inherent_reactions(self):
        return self._has_inherent_reactions()

    def _has_inherent_reactions(self):
        return self._get_parameter_block().has_inherent_reactions

    # Need to separate the existence of inherent reactions from whether they
    # should be included in material balances
    # For some cases, Using an apparent species basis means they can be ignored
    @property
    def include_inherent_reactions(self):
        return self._include_inherent_reactions()

    def _include_inherent_reactions(self):
        return self._get_parameter_block().has_inherent_reactions

    def _get_parameter_block(self):
        # PYLINT-WHY: self._block_data_config_default is set elsewhere,
        # and is supposed to already exist as an attribute by the time
        # (or, it will raise AttributeError at L410)
        # this method is called, so the pylint errors here are false positives
        # pylint: disable=no-member,access-member-before-definition
        try:
            return self._block_data_config_default["parameters"]
        except (KeyError, TypeError):
            # Need to get parameters from initialize dict
            # We will also confirm these are all the same whilst we are at it
            param = None
            if self._block_data_config_default is None:
                self._block_data_config_default = {}
            for v in self._block_data_config_initialize.values():
                if param is None:
                    param = v["parameters"]
                elif param is not v["parameters"]:
                    raise ConfigurationError(
                        "{} StateBlock must use the same parameter block for "
                        "elements. When using the initialize argument, please "
                        "ensure that the same value is used for all parameter "
                        "keys.".format(self.name)
                    )
            self._block_data_config_default["parameters"] = param
        return param

    @property
    def params(self):
        return self._get_parameter_block()

    def fix_initialization_states(self):
        """
        Fixes state variables for state blocks.

        Returns:
            None
        """
        raise NotImplementedError

    def initialize(self, *args, **kwargs):
        """
        This is a default initialization routine for StateBlocks to ensure
        that a routine is present. All StateBlockData classes should
        overload this method with one suited to the particular property package

        Args:
            None

        Returns:
            None
        """
        raise NotImplementedError(
            "{} property package has not implemented an"
            " initialize method. Please contact "
            "the property package developer".format(self.name)
        )

    def report(self, index=(0), true_state=False, dof=False, ostream=None, prefix=""):
        """
        Default report method for StateBlocks. Returns a Block report populated
        with either the display or state variables defined in the
        StateBlockData class.

        Args:
            index : tuple of Block indices indicating which point in time (and
                    space if applicable) to report state at.
            true_state : whether to report the display variables (False
                    default) or the actual state variables (True)
            dof : whether to show local degrees of freedom in the report
                    (default=False)
            ostream : output stream to write report to
            prefix : string to append to the beginning of all output lines

        Returns:
            Printed output to ostream
        """

        if ostream is None:
            ostream = sys.stdout

        # Get DoF and model stats
        if dof:
            dof_stat = degrees_of_freedom(self[index])
            nv = number_variables(self[index])
            nc = number_activated_constraints(self[index])
            nb = number_activated_blocks(self[index])

        # Create stream table
        if true_state:
            disp_dict = self[index].define_state_vars()
        else:
            disp_dict = self[index].define_display_vars()

        stream_attributes = {}

        for k in disp_dict:
            for i in disp_dict[k]:
                if i is None:
                    stream_attributes[k] = disp_dict[k][i]
                else:
                    stream_attributes[k + " " + i] = disp_dict[k][i]

        # Write output
        max_str_length = 84
        tab = " " * 4
        ostream.write("\n" + "=" * max_str_length + "\n")

        lead_str = f"{prefix}State : {self.name}"
        trail_str = f"Index: {index}"
        mid_str = " " * (max_str_length - len(lead_str) - len(trail_str))
        ostream.write(lead_str + mid_str + trail_str)

        if dof:
            ostream.write("\n" + "=" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Local Degrees of Freedom: {dof_stat}")
            ostream.write("\n")
            ostream.write(
                f"{prefix}{tab}Total Variables: {nv}{tab}"
                f"Activated Constraints: {nc}{tab}"
                f"Activated Blocks: {nb}"
            )

        ostream.write("\n" + "-" * max_str_length + "\n")
        ostream.write(f"{prefix}{tab}State Report")

        if any(isinstance(v, VarData) for k, v in stream_attributes.items()):
            ostream.write("\n" * 2)
            ostream.write(f"{prefix}{tab}Variables: \n\n")
            tabular_writer(
                ostream,
                prefix + tab,
                (
                    (k, v)
                    for k, v in stream_attributes.items()
                    if isinstance(v, VarData)
                ),
                ("Value", "Fixed", "Bounds"),
                lambda k, v: ["{:#.5g}".format(value(v)), v.fixed, v.bounds],
            )

        if any(isinstance(v, ExpressionData) for k, v in stream_attributes.items()):
            ostream.write("\n" * 2)
            ostream.write(f"{prefix}{tab}Expressions: \n\n")
            tabular_writer(
                ostream,
                prefix + tab,
                (
                    (k, v)
                    for k, v in stream_attributes.items()
                    if isinstance(v, ExpressionData)
                ),
                ("Value",),
                lambda k, v: ["{:#.5g}".format(value(v))],
            )

        ostream.write("\n" + "=" * max_str_length + "\n")

    def get_port_reference_name(self, component_name, port_name):
        """
        Get the standard name of a "port reference", the component
        accessed on the port by "port_name.component_name".

        Args:
            component_name - name of port member of interest (str)
            port_name - name of Port object (str)

        Returns:
            str with name for Reference used for Port member
        """
        return f"_{component_name}_{port_name}_ref"

    def build_port(
        self,
        doc=None,
        slice_index=None,
        index=None,
    ):
        """
        Constructs a Port based on this StateBlock attached to the target block.

        Args:
            doc - doc string or Prot object
            slice_index - Slice index (e.g. (slice(None), 0.0) that will be
                used to index self when constructing port references. Default = None.
            index - time index to use when calling define_port_members. Default = None.

        Returns:
            Port object and list of tuples with form (Reference, member name)
        """
        if slice_index is None:
            slice_index = Ellipsis
        if index is None:
            index = self.index_set().first()

        # Create empty Port
        p = Port(doc=doc)

        # Get dict of Port members and names
        # Need to get a representative member of StateBlockDatas
        port_member_dict = self[index].define_port_members()

        # Create References for port members
        ref_name_list = []
        for name, component in port_member_dict.items():
            if not component.is_indexed():
                slicer = self[slice_index].component(component.local_name)
            else:
                slicer = self[slice_index].component(component.local_name)[...]

            ref = Reference(slicer)
            ref_name_list.append((ref, name))

            # Add Reference to Port
            p.add(ref, name)

        return p, ref_name_list


class StateBlockData(ProcessBlockData):
    """
    This is the base class for state block data objects. These are
    blocks that contain the Pyomo components associated with calculating a
    set of thermophysical and transport properties for a given material.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "parameters",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="""A reference to an instance of the Property Parameter
Block associated with this property package.""",
        ),
    )
    CONFIG.declare(
        "defined_state",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Flag indicating if incoming state is fully defined",
            doc="""Flag indicating whether the state should be considered fully
defined, and thus whether constraints such as sum of mass/mole fractions should
be included,
**default** - False.
**Valid values:** {
**True** - state variables will be fully defined,
**False** - state variables will not be fully defined.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Phase equilibrium constraint flag",
            doc="""Flag indicating whether phase equilibrium constraints
should be constructed in this state block,
**default** - True.
**Valid values:** {
**True** - StateBlock should calculate phase equilibrium,
**False** - StateBlock should not calculate phase equilibrium.}""",
        ),
    )

    def __init__(self, *args, **kwargs):
        self._lock_attribute_creation = False
        super().__init__(*args, **kwargs)

    def lock_attribute_creation_context(self):
        """Returns a context manager that does not allow attributes to be created
        while in the context and allows attributes to be created normally outside
        the context.
        """
        return _lock_attribute_creation_context(self)

    def is_property_constructed(self, attr):
        """Returns True if the attribute ``attr`` already exists, or false if it
        would be added in ``__getattr__``, or does not exist.

        Args:
            attr (str): Attribute name to check

        Return:
            True if the attribute is already constructed, False otherwise
        """
        with self.lock_attribute_creation_context():
            return hasattr(self, attr)

    @property
    def component_list(self):
        # TODO: Should refactor parent so this is not private
        # pylint: disable-next=protected-access
        return self.parent_component()._return_component_list()

    @property
    def phase_list(self):
        # TODO: Should refactor parent so this is not private
        # pylint: disable-next=protected-access
        return self.parent_component()._return_phase_list()

    @property
    def phase_component_set(self):
        # TODO: Should refactor parent so this is not private
        # pylint: disable-next=protected-access
        return self.parent_component()._return_phase_component_set()

    @property
    def has_inherent_reactions(self):
        # TODO: Should refactor parent so this is not private
        # pylint: disable-next=protected-access
        return self.parent_component()._has_inherent_reactions()

    @property
    def include_inherent_reactions(self):
        # TODO: Should refactor parent so this is not private
        # pylint: disable-next=protected-access
        return self.parent_component()._include_inherent_reactions()

    @property
    def default_initializer(self):
        return self.parent_component().default_initializer

    @property
    def default_scaler(self):
        return self.parent_component().default_scaler

    def build(self):
        """
        General build method for StateBlockDatas.

        Args:
            None

        Returns:
            None
        """
        super(StateBlockData, self).build()
        add_object_reference(self, "_params", self.config.parameters)

    @property
    def params(self):
        return self._params

    def define_state_vars(self):
        """
        Method that returns a dictionary of state variables used in property
        package. Implement a placeholder method which returns an Exception to
        force users to overload this.
        """
        raise NotImplementedError(
            f"{self.name} property package has not implemented the"
            f" define_state_vars method. Please contact "
            f"the property package developer."
        )

    def define_port_members(self):
        """
        Method used to specify components to populate Ports with. Defaults to
        define_state_vars, and developers should overload as required.
        """
        return self.define_state_vars()

    def define_display_vars(self):
        """
        Method used to specify components to use to generate stream tables and
        other outputs. Defaults to define_state_vars, and developers should
        overload as required.
        """
        return self.define_state_vars()

    def get_material_flow_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for material flow to use in
        the material balances.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_material_flow_terms method. Please "
            "contact the property package developer."
        )

    def get_material_density_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for material density to use in
        the material balances .
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_material_density_terms method. Please "
            "contact the property package developer."
        )

    def get_material_diffusion_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for material diffusion to use
        in the material balances.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_material_diffusion_terms method. "
            "Please contact the property package "
            "developer."
        )

    def get_enthalpy_flow_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for enthalpy flow to use in
        the energy balances.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_enthalpy_flow_terms method. Please "
            "contact the property package developer."
        )

    def get_energy_density_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for enthalpy density to use in
        the energy balances.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_energy_density_terms method. Please "
            "contact the property package developer."
        )

    def get_energy_diffusion_terms(self, *args, **kwargs):
        """
        Method which returns a valid expression for energy diffusion to use in
        the energy balances.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " get_energy_diffusion_terms method. "
            "Please contact the property package "
            "developer."
        )

    def get_material_flow_basis(self, *args, **kwargs):
        """
        Method which returns an Enum indicating the basis of the material flow
        term.
        """
        return MaterialFlowBasis.other

    def calculate_bubble_point_temperature(self, *args, **kwargs):
        """
        Method which computes the bubble point temperature for a multi-
        component mixture given a pressure and mole fraction.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " calculate_bubble_point_temperature method."
            " Please contact the property package "
            "developer."
        )

    def calculate_dew_point_temperature(self, *args, **kwargs):
        """
        Method which computes the dew point temperature for a multi-
        component mixture given a pressure and mole fraction.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " calculate_dew_point_temperature method."
            " Please contact the property package "
            "developer."
        )

    def calculate_bubble_point_pressure(self, *args, **kwargs):
        """
        Method which computes the bubble point pressure for a multi-
        component mixture given a temperature and mole fraction.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " calculate_bubble_point_pressure method."
            " Please contact the property package "
            "developer."
        )

    def calculate_dew_point_pressure(self, *args, **kwargs):
        """
        Method which computes the dew point pressure for a multi-
        component mixture given a temperature and mole fraction.
        """
        raise NotImplementedError(
            "{} property package has not implemented the"
            " calculate_dew_point_pressure method."
            " Please contact the property package "
            "developer."
        )

    def __getattr__(self, attr):
        """
        This method is used to avoid generating unnecessary property
        calculations in state blocks. __getattr__ is called whenever a
        property is called for, and if a property does not exist, it looks for
        a method to create the required property, and any associated
        components.

        Create a property calculation if needed. Return an attribute error if
        attr == 'domain' or starts with a _ . The error for _ prevents a
        recursion error if trying to get a function to create a property and
        that function doesn't exist.  Pyomo also occasionally looks for things
        that start with _ and may not exist.  Pyomo also looks for the domain
        attribute, and it may not exist.
        This works by creating a property calculation by calling the "_"+attr
        function.

        A list of __getattr__ calls is maintained in self.__getattrcalls to
        check for recursive loops which maybe useful for debugging. This list
        is cleared after __getattr__ completes successfully.

        Args:
            attr: an attribute to create and return. Should be a property
                  component.
        """
        try:
            # try the Pyomo Block's __getattr__ method first which will return
            # decorators for creating components on the block (e.g. Expression,
            # Constraint, ...).
            return super().__getattr__(attr)
        except AttributeError:
            return build_on_demand(self, attr)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # Get scaling factor defaults, if no scaling factor set
        for v in self.component_data_objects(
            (Constraint, Var, Expression), descend_into=False
        ):
            if iscale.get_scaling_factor(v) is None:  # don't replace if set
                name = v.getname().split("[")[0]
                index = v.index()
                sf = self.config.parameters.get_default_scaling(name, index)
                if sf is not None:
                    iscale.set_scaling_factor(v, sf)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test for Cappresse's module for NMPC.
"""

from pyomo.environ import (
    ConcreteModel,
    SolverFactory,
    units as pyunits,
    TransformationFactory,
)
from pyomo.network import Arc

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.tests.test_initialization import (
    AqueousEnzymeParameterBlock,
    EnzymeReactionParameterBlock,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import CSTR, Mixer, MomentumMixingType
from idaes.core.solvers import get_solver
from idaes.apps.caprese import nmpc
from idaes.apps.caprese.nmpc import *

__author__ = "Robert Parker"


# See if ipopt is available and set up solver
if SolverFactory("ipopt").available():
    solver = get_solver(
        solver="ipopt",
        options={"tol": 1e-6, "mu_init": 1e-8, "bound_push": 1e-8},
    )
else:
    solver = None


def make_model(
    horizon=6, ntfe=60, ntcp=2, inlet_E=11.91, inlet_S=12.92, steady=False, bounds=False
):
    time_set = [0, horizon]

    m = ConcreteModel(name="CSTR model for testing")
    if steady:
        m.fs = FlowsheetBlock(dynamic=False)
    else:
        m.fs = FlowsheetBlock(
            dynamic=True,
            time_set=time_set,
            time_units=pyunits.minute,
        )

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(property_package=m.fs.properties)
    m.fs.cstr = CSTR(
        has_holdup=True,
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        material_balance_type=MaterialBalanceType.componentTotal,
        energy_balance_type=EnergyBalanceType.enthalpyTotal,
        momentum_balance_type=MomentumBalanceType.none,
        has_heat_of_reaction=True,
    )

    m.fs.mixer = Mixer(
        property_package=m.fs.properties,
        material_balance_type=MaterialBalanceType.componentTotal,
        momentum_mixing_type=MomentumMixingType.none,
        num_inlets=2,
        inlet_list=["S_inlet", "E_inlet"],
    )
    # Allegedly the proper energy balance is being used...

    # Time discretization
    if not steady:
        disc = TransformationFactory("dae.collocation")
        disc.apply_to(m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme="LAGRANGE-RADAU")

    # Fix geometry variables
    m.fs.cstr.volume[0].fix(1.0)

    # Fix initial conditions:
    if not steady:
        for p, j in m.fs.properties.phase_list * m.fs.properties.component_list:
            if j == "Solvent":
                continue
            m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0.001)
    # Note: Model does not solve when initial conditions are empty tank

    m.fs.mixer.E_inlet.conc_mol.fix(0)
    m.fs.mixer.S_inlet.conc_mol.fix(0)

    for t, j in m.fs.time * m.fs.properties.component_list:
        if j == "E":
            m.fs.mixer.E_inlet.conc_mol[t, j].fix(inlet_E)
        elif j == "S":
            m.fs.mixer.S_inlet.conc_mol[t, j].fix(inlet_S)

    m.fs.mixer.E_inlet.flow_vol.fix(0.1)
    m.fs.mixer.S_inlet.flow_vol.fix(2.1)
    m.fs.mixer.E_inlet.conc_mol[:, "Solvent"].fix(1.0)
    m.fs.mixer.S_inlet.conc_mol[:, "Solvent"].fix(1.0)

    m.fs.mixer.E_inlet.temperature.fix(290)
    m.fs.mixer.S_inlet.temperature.fix(310)

    m.fs.inlet = Arc(source=m.fs.mixer.outlet, destination=m.fs.cstr.inlet)

    if not steady:
        # This constraint is in lieu of tracking the CSTR's level and allowing
        # the outlet flow rate to be another degree of freedom.
        # ^ Not sure how to do this in IDAES.
        @m.fs.cstr.Constraint(m.fs.time, doc="Total flow rate balance")
        def total_flow_balance(cstr, t):
            return cstr.inlet.flow_vol[t] == cstr.outlet.flow_vol[t]

    # Specify initial condition for energy
    if not steady:
        m.fs.cstr.control_volume.energy_holdup[m.fs.time.first(), "aq"].fix(300)

    TransformationFactory("network.expand_arcs").apply_to(m.fs)

    if bounds:
        m.fs.mixer.E_inlet.flow_vol.setlb(0.01)
        m.fs.mixer.E_inlet.flow_vol.setub(1.0)
        m.fs.mixer.S_inlet.flow_vol.setlb(0.5)
        m.fs.mixer.S_inlet.flow_vol.setub(5.0)

        m.fs.cstr.control_volume.material_holdup.setlb(0)
        holdup = m.fs.cstr.control_volume.material_holdup
        for t in m.fs.time:
            holdup[t, "aq", "S"].setub(20)
            holdup[t, "aq", "E"].setub(1)
            holdup[t, "aq", "P"].setub(5)
            holdup[t, "aq", "C"].setub(5)

    return m


if __name__ == "__main__":
    m_plant = make_model()
    m_controller = make_model(bounds=True)
    m_ss = make_model(steady=True)

    assert degrees_of_freedom(m_plant) == 0
    assert degrees_of_freedom(m_ss) == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains tests for the NominalValueExtractionVisitor.
"""
import math
import logging
import pytest

import pyomo.environ as pyo
from pyomo.contrib.pynumero.asl import AmplInterface

from idaes.core.scaling.util import NominalValueExtractionVisitor, set_scaling_factor
from idaes.models.properties.modular_properties.eos.ceos_common import (
    cubic_roots_available,
    CubicThermoExpressions,
    CubicType as CubicEoS,
)

__author__ = "Andrew Lee"


class TestNominalValueExtractionVisitor:
    @pytest.fixture(scope="class")
    def m(self):
        m = pyo.ConcreteModel()
        m.set = pyo.Set(initialize=["a", "b", "c"])

        return m

    @pytest.mark.unit
    def test_int(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=7) == [7]

    @pytest.mark.unit
    def test_float(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=7.7) == [7.7]

    @pytest.mark.unit
    def test_negative_float(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=-7.7) == [-7.7]

    @pytest.mark.unit
    def test_zero(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=0) == [0]

    @pytest.mark.unit
    def test_true(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=True) == [1]

    @pytest.mark.unit
    def test_false(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=False) == [0]

    @pytest.mark.unit
    def test_scalar_param_no_scale(self, m):
        m.scalar_param = pyo.Param(initialize=1, mutable=True)
        assert NominalValueExtractionVisitor().walk_expression(expr=m.scalar_param) == [
            1
        ]

    @pytest.mark.unit
    def test_scalar_param_w_scale(self, m):
        m.scalar_param = pyo.Param(default=12, mutable=True)
        set_scaling_factor(m.scalar_param, 1 / 10)
        assert NominalValueExtractionVisitor().walk_expression(expr=m.scalar_param) == [
            12
        ]

    @pytest.mark.unit
    def test_indexed_param_w_scale(self, m):
        m.indexed_param = pyo.Param(m.set, initialize=1, mutable=True)
        set_scaling_factor(m.indexed_param["a"], 1 / 13)
        set_scaling_factor(m.indexed_param["b"], 1 / 14)
        set_scaling_factor(m.indexed_param["c"], 1 / 15)

        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_param["a"]
        ) == [1]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_param["b"]
        ) == [1]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_param["c"]
        ) == [1]

    @pytest.mark.unit
    def test_scalar_var_no_scale(self, m):
        m.scalar_var = pyo.Var(initialize=10)
        # Should use current value
        assert NominalValueExtractionVisitor().walk_expression(expr=m.scalar_var) == [
            10
        ]

    @pytest.mark.unit
    def test_scalar_var_w_scale(self, m):
        set_scaling_factor(m.scalar_var, 1 / 21)
        assert NominalValueExtractionVisitor().walk_expression(expr=m.scalar_var) == [
            21
        ]

    @pytest.mark.unit
    def test_var_neg_bounds(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(bounds=(-1000, 0))
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be negative
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [-4]

    @pytest.mark.unit
    def test_var_neg_upper_bound(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(bounds=(None, -2000))
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be negative
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [-4]

    @pytest.mark.unit
    def test_var_neg_domain(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(domain=pyo.NegativeReals)

        set_scaling_factor(m.var, 1 / 4)
        # Expect nominal value to be negative
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [-4]

    @pytest.mark.unit
    def test_var_neg_value(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(initialize=-1)
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be negative
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [-4]

    @pytest.mark.unit
    def test_var_fixed_value(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(initialize=-1)
        m.var.fix()
        set_scaling_factor(m.var, 1 / 4)

        # Nominal value should be value
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [-1]

    @pytest.mark.unit
    def test_var_pos_bounds(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(bounds=(0, 1000))
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be positive
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [4]

    @pytest.mark.unit
    def test_var_pos_lower_bound(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(bounds=(1000, None))
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be positive
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [4]

    @pytest.mark.unit
    def test_var_pos_domain(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(domain=pyo.PositiveReals)
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be positive
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [4]

    @pytest.mark.unit
    def test_var_pos_value(self):
        m = pyo.ConcreteModel()
        m.var = pyo.Var(initialize=1)
        set_scaling_factor(m.var, 1 / 4)

        # Expect nominal value to be positive
        assert NominalValueExtractionVisitor().walk_expression(expr=m.var) == [4]

    @pytest.mark.unit
    def test_indexed_var_no_scale(self, m):
        m.indexed_var = pyo.Var(m.set, initialize=1)
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["a"]
        ) == [1]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["b"]
        ) == [1]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["c"]
        ) == [1]

    @pytest.mark.unit
    def test_indexed_var_w_scale(self, m):
        set_scaling_factor(m.indexed_var["a"], 1 / 22)
        set_scaling_factor(m.indexed_var["b"], 1 / 23)
        set_scaling_factor(m.indexed_var["c"], 1 / 24)

        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["a"]
        ) == [22]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["b"]
        ) == [23]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["c"]
        ) == [24]

    @pytest.mark.unit
    def test_indexed_var_w_scale_partial_fixed(self, m):
        m.indexed_var["a"].fix(20)
        set_scaling_factor(m.indexed_var["a"], 1 / 22)
        set_scaling_factor(m.indexed_var["b"], 1 / 23)
        set_scaling_factor(m.indexed_var["c"], 1 / 24)

        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["a"]
        ) == [20]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["b"]
        ) == [23]
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.indexed_var["c"]
        ) == [24]

        # Clean up for future tests
        m.indexed_var["a"].unfix()

    @pytest.mark.unit
    def test_equality_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var == m.indexed_var["a"]
        ) == [21, 22]

    @pytest.mark.unit
    def test_inequality_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var <= m.indexed_var["a"]
        ) == [21, 22]

    @pytest.mark.unit
    def test_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=sum(m.indexed_var[i] for i in m.set)
        ) == [22, 23, 24]

    @pytest.mark.unit
    def test_additive_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var + m.indexed_var["a"] + m.scalar_param
        ) == [21, 22, 12]

    @pytest.mark.unit
    def test_additive_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var + m.indexed_var["a"] - m.scalar_param
        ) == [21, 22, -12]

    @pytest.mark.unit
    def test_product_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var * m.indexed_var["a"] * m.scalar_param
        ) == [21 * 22 * 12]

    @pytest.mark.unit
    def test_product_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var + m.indexed_var["a"])
            * (m.scalar_param + m.indexed_var["b"])
        ) == [21 * 12, 21 * 23, 22 * 12, 22 * 23]

    @pytest.mark.unit
    def test_product_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var - m.indexed_var["a"])
            * (m.scalar_param - m.indexed_var["b"])
        ) == [21 * 12, -21 * 23, -22 * 12, 22 * 23]

    @pytest.mark.unit
    def test_division_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var / m.indexed_var["a"] / m.scalar_param
        ) == [21 / 22 / 12]

    @pytest.mark.unit
    def test_division_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var + m.indexed_var["a"])
            / (m.scalar_param + m.indexed_var["b"])
        ) == [(21 + 22) / (12 + 23)]

    @pytest.mark.unit
    def test_division_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var - m.indexed_var["a"])
            / (m.scalar_param - m.indexed_var["b"])
        ) == [(21 - 22) / (12 - 23)]

    @pytest.mark.unit
    def test_division_expr_error(self, m, caplog):
        caplog.set_level(logging.DEBUG, logger="idaes.core.scaling")
        assert NominalValueExtractionVisitor().walk_expression(
            expr=1 / (m.scalar_var - 21)
        ) == [1]

        expected = "Nominal value of 0 found in denominator of division expression. "
        "Assigning a value of 1. You should check you scaling factors and models to "
        "ensure there are no values of 0 that can appear in these functions."

        assert expected in caplog.text

    @pytest.mark.unit
    def test_pow_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.scalar_var ** m.indexed_var["a"]
        ) == pytest.approx([21**22], rel=1e-12)

    @pytest.mark.unit
    def test_pow_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var + m.indexed_var["a"])
            ** (m.scalar_param + m.indexed_var["b"])
        ) == [
            pytest.approx((21 + 22) ** (12 + 23), rel=1e-12),
        ]

    @pytest.mark.unit
    def test_pow_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=(m.scalar_var - m.indexed_var["a"])
            ** (m.scalar_param - m.indexed_var["b"])
        ) == [
            pytest.approx(abs(21 - 22) ** (12 - 23), rel=1e-12),
        ]

    @pytest.mark.unit
    def test_negation_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=-m.scalar_var) == [
            -21
        ]

    @pytest.mark.unit
    def test_negation_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=-(m.scalar_var + m.indexed_var["a"])
        ) == [-21, -22]

    @pytest.mark.unit
    def test_log_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log(m.scalar_var)
        ) == [pytest.approx(math.log(21), rel=1e-12)]

    @pytest.mark.unit
    def test_log_expr_error(self, m):
        with pytest.raises(
            ValueError,
            match="Evaluation error occurred when getting nominal value in log expression "
            "with input 0.0. You should check you scaling factors and model to "
            "address any numerical issues or scale this constraint manually.",
        ):
            assert NominalValueExtractionVisitor().walk_expression(
                expr=pyo.log(m.scalar_var - 21)
            )

    @pytest.mark.unit
    def test_log_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.log(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_log_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log(-m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.log(-21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_log10_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log10(m.scalar_var)
        ) == [pytest.approx(math.log10(21), rel=1e-12)]

    @pytest.mark.unit
    def test_log10_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log10(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.log10(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_log10_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.log10(-m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.log10(-21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_log10_expr_error(self, m):
        with pytest.raises(
            ValueError,
            match="Evaluation error occurred when getting nominal value in log10 expression "
            "with input 0.0. You should check you scaling factors and model to "
            "address any numerical issues or scale this constraint manually.",
        ):
            assert NominalValueExtractionVisitor().walk_expression(
                expr=pyo.log10(m.scalar_var - 21)
            )

    @pytest.mark.unit
    def test_sqrt_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sqrt(m.scalar_var)
        ) == [pytest.approx(21**0.5, rel=1e-12)]

    @pytest.mark.unit
    def test_sqrt_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sqrt(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx((21 + 22) ** 0.5, rel=1e-12)]

    @pytest.mark.unit
    def test_sqrt_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sqrt(-m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx((-21 + 22) ** 0.5, rel=1e-12)]

    @pytest.mark.unit
    def test_sqrt_expr_error(self, m):
        with pytest.raises(
            ValueError,
            match="Evaluation error occurred when getting nominal value in sqrt expression "
            "with input -21.0. You should check you scaling factors and model to "
            "address any numerical issues or scale this constraint manually.",
        ):
            assert NominalValueExtractionVisitor().walk_expression(
                expr=pyo.sqrt(-m.scalar_var)
            )

    @pytest.mark.unit
    def test_sin_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sin(m.scalar_var)
        ) == [pytest.approx(math.sin(21), rel=1e-12)]

    @pytest.mark.unit
    def test_sin_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sin(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.sin(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_sin_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sin(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.sin(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_cos_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cos(m.scalar_var)
        ) == [pytest.approx(math.cos(21), rel=1e-12)]

    @pytest.mark.unit
    def test_cos_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cos(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.cos(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_cos_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cos(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.cos(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_tan_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tan(m.scalar_var)
        ) == [pytest.approx(math.tan(21), rel=1e-12)]

    @pytest.mark.unit
    def test_tan_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tan(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.tan(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_tan_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tan(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.tan(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_sinh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sinh(m.scalar_var)
        ) == [pytest.approx(math.sinh(21), rel=1e-12)]

    @pytest.mark.unit
    def test_sinh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sinh(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.sinh(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_sinh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.sinh(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.sinh(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_cosh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cosh(m.scalar_var)
        ) == [pytest.approx(math.cosh(21), rel=1e-12)]

    @pytest.mark.unit
    def test_cosh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cosh(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.cosh(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_cosh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.cosh(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.cosh(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_tanh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tanh(m.scalar_var)
        ) == [pytest.approx(math.tanh(21), rel=1e-12)]

    @pytest.mark.unit
    def test_tanh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tanh(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.tanh(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_tanh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.tanh(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.tanh(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_asin_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(expr=pyo.asin(1)) == [
            pytest.approx(math.asin(1), rel=1e-12)
        ]

    @pytest.mark.unit
    def test_asin_sum_expr(self, m):
        m.scalar_param.set_value(0.5)
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.asin(0.5 + m.scalar_param)
        ) == [pytest.approx(math.asin(1), rel=1e-12)]

    @pytest.mark.unit
    def test_asin_sum_expr_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.asin(0.5 - m.scalar_param)
        ) == [pytest.approx(math.asin(0), rel=1e-12)]

    @pytest.mark.unit
    def test_acos_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acos(m.scalar_param)
        ) == [pytest.approx(math.acos(0.5), rel=1e-12)]

    @pytest.mark.unit
    def test_acos_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acos(0.5 + m.scalar_param)
        ) == [pytest.approx(math.acos(1), rel=1e-12)]

    @pytest.mark.unit
    def test_acos_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acos(0.5 - m.scalar_param)
        ) == [pytest.approx(math.acos(0), rel=1e-12)]

    @pytest.mark.unit
    def test_asinh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.asinh(m.scalar_var)
        ) == [pytest.approx(math.asinh(21), rel=1e-12)]

    @pytest.mark.unit
    def test_asinh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.asinh(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.asinh(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_asinh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.asinh(m.scalar_var - m.indexed_var["a"])
        ) == [pytest.approx(math.asinh(21 - 22), rel=1e-12)]

    @pytest.mark.unit
    def test_acosh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acosh(m.scalar_var)
        ) == [pytest.approx(math.acosh(21), rel=1e-12)]

    @pytest.mark.unit
    def test_acosh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acosh(m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.acosh(21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_acosh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.acosh(-m.scalar_var + m.indexed_var["a"])
        ) == [pytest.approx(math.acosh(-21 + 22), rel=1e-12)]

    @pytest.mark.unit
    def test_atanh_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.atanh(m.scalar_param)
        ) == [pytest.approx(math.atanh(0.5), rel=1e-12)]

    @pytest.mark.unit
    def test_atanh_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.atanh(0.4 + m.scalar_param)
        ) == [pytest.approx(math.atanh(0.9), rel=1e-12)]

    @pytest.mark.unit
    def test_atanh_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.atanh(0.4 - m.scalar_param)
        ) == [pytest.approx(math.atanh(-0.1), rel=1e-12)]

    @pytest.mark.unit
    def test_exp_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.exp(m.scalar_param)
        ) == [pytest.approx(math.exp(0.5), rel=1e-12)]

    @pytest.mark.unit
    def test_exp_sum_expr(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.exp(0.4 + m.scalar_param)
        ) == [pytest.approx(math.exp(0.9), rel=1e-12)]

    @pytest.mark.unit
    def test_exp_sum_expr_w_negation(self, m):
        assert NominalValueExtractionVisitor().walk_expression(
            expr=pyo.exp(-0.4 + m.scalar_param)
        ) == [pytest.approx(math.exp(0.1), rel=1e-12)]

    @pytest.mark.unit
    def test_expr_if(self, m):
        m.exprif = pyo.Expr_if(
            IF=m.scalar_param,
            THEN=m.indexed_var["a"],
            ELSE=m.indexed_var["b"] + m.indexed_var["c"],
        )

        assert NominalValueExtractionVisitor().walk_expression(expr=m.exprif) == [
            22,
            23,
            24,
        ]

    @pytest.mark.unit
    def test_expr_if_w_negation(self, m):
        m.exprif = pyo.Expr_if(
            IF=m.scalar_param,
            THEN=m.indexed_var["a"],
            ELSE=m.indexed_var["b"] - m.indexed_var["c"],
        )

        assert NominalValueExtractionVisitor().walk_expression(expr=m.exprif) == [
            22,
            23,
            -24,
        ]

    @pytest.mark.unit
    @pytest.mark.skipif(
        not AmplInterface.available(), reason="pynumero_ASL is not available"
    )
    @pytest.mark.skipif(not cubic_roots_available, reason="Cubic roots not available")
    def test_ext_func(self):
        # Use the cubic root external function to test
        m = pyo.ConcreteModel()
        m.a = pyo.Var(initialize=1)
        m.b = pyo.Var(initialize=1)

        set_scaling_factor(m.a, 1 / 2)
        set_scaling_factor(m.b, 1 / 4)

        m.expr_write = CubicThermoExpressions(m)
        Z = m.expr_write.z_liq(eos=CubicEoS.PR, A=m.a, B=m.b)

        expected_mag = -9.489811292072448
        assert NominalValueExtractionVisitor().walk_expression(expr=Z) == [
            pytest.approx(expected_mag, rel=1e-8)
        ]

        # Check that model state did not change
        assert pyo.value(m.a) == 1
        assert pyo.value(m.b) == 1
        assert pyo.value(Z) == pytest.approx(-2.1149075414767577, rel=1e-8)

        # Now, change the actual state to the expected magnitudes and confirm result
        m.a.set_value(2)
        m.b.set_value(4)
        assert pyo.value(Z) == pytest.approx(expected_mag, rel=1e-8)

    @pytest.mark.unit
    def test_Expression(self, m):
        m.expression = pyo.Expression(
            expr=m.scalar_param ** (sum(m.indexed_var[i] for i in m.set))
        )

        assert NominalValueExtractionVisitor().walk_expression(expr=m.expression) == [
            0.5 ** (22 + 23 + 24)
        ]

    @pytest.mark.unit
    def test_constraint(self, m):
        m.constraint = pyo.Constraint(expr=m.scalar_var == m.expression)

        assert NominalValueExtractionVisitor().walk_expression(
            expr=m.constraint.expr
        ) == [21, 0.5 ** (22 + 23 + 24)]
